#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 Axigen Messaging SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Definitions of classes related to CLI Contexts
    Context
      RootContext
      ServerContext
      DomainContext
      AccountContext
    ContextStack
"""

__copyright__ = "Copyright (C) 2020, Axigen Messaging SRL"
__license__ = "Apache License Version 2.0"
__version__ = "1.1.0"
__author__ = 'eugen.belea@axigen.com (Eugen Belea)'
__maintainer__ = 'eugen.belea@axigen.com (Eugen Belea)'


from . import clierror

class Context():
    '''
    CLI Context (base class of all CLI contexts)
    '''
    def __init__(self, _name="", _parentType=None, _hasCommit=False, _hasBack=False):
        self.name = _name
        self.parentType = _parentType
        self.hasCommit = _hasCommit
        self.hasBack = _hasBack


class ContextStack():
    '''
    CLI Context Stack (existing on a CLI connection)
    '''
    def __init__(self, _log=None):
        self.stack = []
        self.log = _log
    
    def isEmpty(self):
        return len(self.stack) == 0;
        
    def push(self, ctx):
        self.stack.append(ctx)
        if self.log:
            self.log.info("context push (%u) '<%s>'", \
                len(self.stack), ctx)
    
    def front(self):
        if self.isEmpty():
            raise clierror.InvalidContextError()
        return self.stack[-1]
    
    def pop(self):
        if self.isEmpty():
            raise clierror.InvalidContextError()
        ctx = self.stack.pop()
        if self.log:
            self.log.info("context pop (%u) '<%s>'", \
                len(self.stack), ctx)
        return ctx
    
class RootContext(Context):
    '''
    CLI Root Context
    '''
    def __init__(self):
        super.__init__(self, "<#>")


class ServerContext(Context):
    '''
    CLI Server Context
    '''
    def __init__(self):
        super.__init__(self, "<server#>", RootContext, True, True)


class DomainContext(Context):
    '''
    CLI Domain Context
    '''
    def __init__(self, _domName):
        super.__init__(self, "<domain#>", RootContext, True, True)
        self.domName = _domName        


class AccountContext(Context):
    '''
    CLI Account Context
    '''
    def __init__(self, _acctName):
        super.__init__(self, "<domain-account#>", DomainContext, True, True)
        self.acctName = _acctName        

