
from __future__ import print_function

"""
Axigen backup module.
"""

import re
import os
import stat
from . import bkparser
import traceback
import sys
import time

class ProgressIndicator:
    """
    Abstract class used to display the progress of the backup operation.
    """
    def __init__(self):
        pass
    
    def beginBackup(self, srcRoot, destRoot, restoreMode):
        """
        Called to report the begining of the backup operation
        """
        pass
    def endBackup(self, successfull, copyFileCount, copyFileSize):
        """
        Called to report the end of the backup operation
        """
        pass
    def abortBackup(self, copyFileCount, copyFileSize):
        """
        Called to report the premature end of the backup operation at user's request
        """
    def reportGlobalProgress(self, backupProgress, backupCount, copyFileCount, copyFileSize):
        """
        Called to report some progress on the global backup progress bar
        """
        pass
    def reportError(self, errorDetails):
        """
        Called to report some error
        """
        pass        
    def beginDomainBackup(self, backupDomain, domainIdx, domainCount):
        """
        Called when backup operation is started for a domain.
        """
        pass
    def endDomainBackup(self, backupDomain, successfull, domainIdx, domainCount):
        """
        Called when backup operation is finalized for a domain.
        """
        pass
    def beginDomainObjectBackup(self, backupDomainObject, domainObjectIdx, domainObjectCount):
        """
        Called when backup operation is started for a domain object.
        """
        pass
    def endDomainObjectBackup(self, backupDomainObject, successfull, domainObjectIdx, domainObjectCount):
        """
        Called when backup operation is finalized for a domain object.
        """
        pass
    def beginFolderBackup(self, backupFolder, folderIdx, folderCount):
        """
        Called when backup operation is started for a folder.
        """
        pass
    def endFolderBackup(self, backupFolder, successfull, folderIdx, folderCount):
        """
        Called when backup operation is finalized for a folder.
        """
        pass
    def beginMailBackup(self, backupFolder, mailIdx, mailCount):
        """
        Called when backup operation is started for a mail.
        """
        pass
    def endMailBackup(self, backupFolder, successfull, mailIdx, mailCount):
        """
        Called when backup operation is finalized for a mail.
        """
        pass

def _copyFileInternal(srcFileName, destFileName, removeFile=True):
    srcFile = None
    destFile = None
    rval = -1
    toWrite = 0
    copySize = 0
    keyboardInterrupt = False
    destFileMode = 0
    try:
        if hasattr(os, "O_BINARY"):
            srcFile = os.open(srcFileName, os.O_RDONLY | os.O_BINARY)
            destFileMode = destFileMode | os.O_BINARY
        else:
            srcFile = os.open(srcFileName, os.O_RDONLY)

        # get source file info
        srcFileStatInfo = os.stat(srcFileName)

        # OTRS-73711-66 : nothing to do when source is empty and destination file is missing
        if (srcFileStatInfo.st_size == 0) and (not os.path.isfile(destFileName)):
            if os.getenv('USE_DEBUG', False):
                print("Empty file detected: " + str(srcFileName))
            # OTRS-77987 : we have to close empty files
            os.close(srcFile)
            srcFile = None
            return 0
        
        # if we want to keep old removed files don't overwrite their content by opening with O_TRUNC
        if (not removeFile) and (srcFileStatInfo.st_size == 0):
            destFileMode = destFileMode | os.O_RDONLY | os.O_CREAT
        else:
            destFileMode = destFileMode | os.O_WRONLY | os.O_CREAT | os.O_TRUNC
        destFile = os.open(destFileName, destFileMode, 0o666)
        
        removeDestFile = True
        while True:
            data = os.read(srcFile, 16384)
            if not data:
                break
            removeDestFile = False
            toWrite = len(data)
            copySize = copySize + toWrite
            while toWrite > 0:
                written = os.write(destFile, data)
                if written < 1:
                    break
                toWrite = toWrite - written
                data = data[written:]
            if toWrite > 0:
                break
        os.close(destFile)
        destFile = None
        if toWrite == 0:
            if removeDestFile and removeFile:
                try:
                    if os.getenv('USE_DEBUG', False):
                      print("Deleting " + str(destFileName))
                    os.remove(destFileName)
                    if os.getenv('USE_DEBUG', False):
                      print("Deleted: " + str(destFileName))
                except:
                    if os.getenv('USE_DEBUG', False):
                      print("Delete error for: {0} = {1} | {2}".format(destFileName, sys.exc_info()[0], sys.exc_info()[1]))
                      traceback.print_tb(sys.exc_info()[2], limit=5, file=sys.stdout)
                    pass
            rval = copySize
        os.close(srcFile)
        srcFile = None
        # RM3408 : Get stat info from srcFileName and set time attributes to destFileName
        # OTRS-71719-26 : but not when we like to keep the removed files
        if srcFileStatInfo.st_size > 0:
            try:
                os.utime(destFileName,(srcFileStatInfo.st_atime,srcFileStatInfo.st_mtime))
            except:
                pass
    except KeyboardInterrupt:
        keyboardInterrupt = True
    except:
        if os.getenv('USE_DEBUG', False):
          print("Copy error: {0} -> {1} = {2} | {3}".format(srcFileName, destFileName, sys.exc_info()[0], sys.exc_info()[1]))
          traceback.print_tb(sys.exc_info()[2], limit=5, file=sys.stdout)
        pass
    if srcFile != None:
        try:
            os.close(srcFile)
        except:
            pass
    if destFile != None:
        try:
            os.close(destFile)
            # RM3408 : Get stat info from srcFileName and set time attributes to destFileName
            # OTRS-71719-26 : but not when we like to keep the removed files
            if srcFileStatInfo.st_size > 0:
                try:
                    os.utime(destFileName,(srcFileStatInfo.st_atime,srcFileStatInfo.st_mtime))
                except:
                    pass
        except:
            pass

    if keyboardInterrupt:
        raise KeyboardInterrupt()
    return rval

def copyFile(srcFileName, destFileName, removeFile=True):
    for retry in range(0, 5):
        rval = _copyFileInternal(srcFileName, destFileName, removeFile)
        if(rval >= 0):
            return rval
        #print("Copyfile ({0} -> {1}) error; retry".format(srcFileName, destFileName))
        time.sleep(1)
    return _copyFileInternal(srcFileName, destFileName, removeFile)

def listDir(dirPath):
    for retry in range(0, 5):
        try:
            return os.listdir(dirPath)
        except KeyboardInterrupt:
            raise
        except:
            #print("Listdir({0}) error ({1} | {2}); retry".format(dirPath, sys.exc_info()[0], sys.exc_info()[1]))
            time.sleep(1)
    return os.listdir(dirPath)

def deleteObject(objectPath, removeObjectDir = True):
    try:
        for root, dirs, files in os.walk(objectPath, topdown = False):
            for name in files:
                os.remove(os.path.join(root, name))
            for name in dirs:
                os.rmdir(os.path.join(root, name))
        if removeObjectDir:
            os.rmdir(objectPath)
        return True
    except KeyboardInterrupt:
        raise
    except:
        return False
    
class Engine:
    """
    The backup engine used to synchronize a destination backup folder
    with a source backup folder
    """
    def __init__(self, srcPath, destPath):
        self.backupSourcePath = srcPath
        self.backupDestinationPath = destPath
        self.deleteRemovedDomains = False
        self.deleteRemovedDomainObjects = False
        self.deleteRemovedFolders = True
        self.deleteRemovedMails = True
        self.debugMode = os.getenv('USE_DEBUG', False)
        self._copyFileCount = 0
        self._copyFileSize = 0
        self._srcRoot = bkparser.BackupRoot(srcPath)
        self._destRoot = bkparser.BackupRoot(destPath)
        self._reporters = []
        self._gobjCount = 0
        self._gobjIdx = 0
        self._restrictDomain = None
        self._restrictDomainObject = None
        self._restrictFolder = None
        self._keyboardInterrupt = False
        self._restoreMode = None
        self._restoreNewDomain = None
        self._restoreNewDomainObject = None
        self._restoreNewFolder = None
        self._nextVirtualId = bkparser.virtualObjectId
        
    def registerProgressIndicator(self, progressIndicator):
        self._reporters.append(progressIndicator)

    def restrictBackup(self, restrictDomain = None, restrictDomainObject = None, restrictFolder = None):
        self._restrictDomain = restrictDomain
        self._restrictDomainObject = restrictDomainObject
        self._restrictFolder = restrictFolder
        
    def backup(self):
        """
        Synchronize two backup directories
        """
        self._keyboardInterrupt = False
        self._restoreMode = False
        self._beginBackup()        
        try:
            rval = self._doBackup()
        except:
            self._reportError("Unexpected error during backup: {0} | {1}".format(sys.exc_info()[0], sys.exc_info()[1]))
            if os.getenv('USE_DEBUG', False):
                traceback.print_tb(sys.exc_info()[2], limit=10, file=sys.stdout)
            rval = False
        if (not rval) and self._keyboardInterrupt:
            self._abortBackup()
        else:
            self._endBackup(rval)
        return rval

    def restore(self):
        """
        Synchronize two backup directories in restore mode
        """
        self._keyboardInterrupt = False
        self._restoreMode = True
        self.deleteRemovedDomains = False
        self.deleteRemovedDomainObjects = False
        self.deleteRemovedFolders = False
        
        self._beginBackup()
        try:
            rval = self._doBackup()
        except:
            self._reportError("Unexpected error during restore: {0} | {1}".format(sys.exc_info()[0], sys.exc_info()[1]))
            if os.getenv('USE_DEBUG', False):
                traceback.print_tb(sys.exc_info()[2], limit=10, file=sys.stdout)
            rval = False
        if (not rval) and self._keyboardInterrupt:
            self._abortBackup()
        else:
            self._endBackup(rval)
        return rval

    def _doBackup(self):
        # check if source is a directory
        try:
            if not stat.S_ISDIR(os.stat(self.backupSourcePath).st_mode):
                self._reportError("Source path is not an axigen domains directory")
                return False
        except:
            self._reportError("Source path is not an axigen domains directory")
            return False

        # check if destination is a directory
        try:
            if not stat.S_ISDIR(os.stat(self.backupDestinationPath).st_mode):
                self._reportError("Destination path is not a directory")
                return False
        except:
            try:
                os.makedirs(self.backupDestinationPath)
            except:
                self._reportError("Unable to create destination directory")
                return False

        # scan source and destination directory
        try:
            if not self._srcRoot.scan():
                self._reportError("Unable to scan source directory")
                return False
        except:
            self._reportError("Error scanning source directory")
            return False
        try:
            if not self._destRoot.scan():
                self._reportError("Unable to scan destination directory")
                return False
        except:
            self._reportError("Error scanning destination directory")
            return False

        srcDomains=set(self._srcRoot.listIds())
        destDomains=set(self._destRoot.listIds())
		
        if len(srcDomains) == 0:
            self._reportError("No domains found in source directory")
            return True
		
        if self.deleteRemovedDomains:
            removeDomains = destDomains - srcDomains
        else:
            removeDomains = set()
        addDomains = srcDomains - destDomains
        updateDomains = srcDomains & destDomains
        # compute global backup object count and domain count
        domainCount = len(removeDomains) + len(srcDomains)
        self._gobjCount = domainCount + sum(self._srcRoot.listDomainObjectsCount()) + len(srcDomains)
        if self.deleteRemovedDomainObjects:
            for domainId in updateDomains:
                srcDom = self._srcRoot.getDomainById(domainId)
                destDom = self._destRoot.getDomainById(domainId)
                if (srcDom == None) or (destDom == None):
                    continue
                self._gobjCount = self._gobjCount + len(set(destDom.listIds()) - set(srcDom.listIds()))
        self._gobjIdx = 0
        self._reportGlobalProgress()
        domainIdx = 1
        domainsOk = True

        # start backup/restore
        # first remove domains (if any)
        for domainId in removeDomains:
            if self._keyboardInterrupt:
                return False
            destDom = self._destRoot.getDomainById(domainId)            
            if destDom != None:
                if (self._restrictDomain != None) and (self._restrictDomain != destDom.getName()):
                    continue
                self._beginDomainBackup(destDom, domainIdx, domainCount)
                domainSuccess = self._deleteObject(destDom.getPath())
                self._endDomainBackup(destDom, domainSuccess, domainIdx, domainCount)
            self._reportGlobalProgress()
            domainIdx = domainIdx + 1

        # now backup/restore source domains
        for domainId in srcDomains:
            if self._keyboardInterrupt:
                return False
            srcDom = self._srcRoot.getDomainById(domainId)
            if srcDom != None:
                if (self._restrictDomain != None) and (self._restrictDomain != srcDom.getName()):
                    continue
                self._beginDomainBackup(srcDom, domainIdx, domainCount)
                domainSuccess = False
                try:
                    if self._restoreMode:
                        destDom = self._destRoot.getDomainById(domainId)
                        if destDom == None:
                            self._restoreNewDomain = True
                            destDom = self._destRoot.getDomainByName(srcDom.getName())
                        else:
                            self._restoreNewDomain = False
                        if destDom != None:
                            domainSuccess = self._doDomainBackup(srcDom, destDom)
                        else:
                            self._reportError("Destination domain '{0}' not found".format(srcDom.getName()))
                    else:
                        if self._destRoot.setDomain(domainId, srcDom.getName()):
                            self._restoreNewDomain = False
                            destDom = self._destRoot.getDomainById(domainId)
                            if destDom != None:
                                domainSuccess = self._doDomainBackup(srcDom, destDom)
                            else:
                                self._reportError("Destination domain '{0}' not found".format(srcDom.getName()))
                        else:
                            self._reportError("Error creating domain '{0}' in destination directory".format(srcDom.getName()))
                except KeyboardInterrupt:
                    self._keyboardInterrupt = True
                except:
                    if self._restoreMode:
                        self._reportError("Unexpected error while restoring domain '{0}': {1} | {2}".format(srcDom.getName(), sys.exc_info()[0], sys.exc_info()[1]))
                    else:
                        self._reportError("Unexpected error while backup domain '{0}': {1} | {2}".format(srcDom.getName(), sys.exc_info()[0], sys.exc_info()[1]))
                    if os.getenv('USE_DEBUG', False):
                        traceback.print_tb(sys.exc_info()[2], limit=10, file=sys.stdout)    
                self._restoreNewDomain = None
                domainsOk = domainsOk and domainSuccess
                self._endDomainBackup(srcDom, domainSuccess, domainIdx, domainCount)
            self._reportGlobalProgress()
            domainIdx = domainIdx + 1
        return domainsOk

    def _doDomainBackup(self, srcDom, destDom):
        if self._restrictDomainObject == None:
            # copy domain config files
            for fname in listDir(srcDom.getPath()):
                if fname[-4:] == ".abf":
                    if not self._copyFile(os.path.join(srcDom.getPath(), fname), os.path.join(destDom.getPath(), fname)):
                        self._reportError("Unable to copy domain '{0}' backup file '{1}'".format(srcDom.getName(), fname))
                        return False

        if not srcDom.scan():
            if destDom.scan():
                self._reportError("Incompatible backup domain '{0}'".format(srcDom.getName()))
                return False
            return True
        if not destDom.scan():
            self._reportError("Incompatible backup domain '{0}'".format(srcDom.getName()))
            return False

        # start backup/restore domain objects
        objectIdx = 1
        objectCount = len(srcDom.listIds())
        objectsOk = True
        if self.deleteRemovedDomainObjects:
            removedList = set(destDom.listIds()) - set(srcDom.listIds())
            objectCount = objectCount + len(removedList)
            for objId in removedList:
                if self._keyboardInterrupt:
                    return False
                destObj = destDom.getObject(objId)
                if destObj != None:
                    if (self._restrictDomainObject != None) and (self._restrictDomainObject != destObj.getName()):
                        continue
                    self._beginDomainObjectBackup(destObj, objectIdx, objectCount)
                    objectSuccess = self._deleteObject(destObj.getPath())                    
                    self._endDomainBackup(destObj, objectSuccess, objectIdx, objectCount)
                self._reportGlobalProgress()
                objectIdx = objectIdx + 1

        # now backup/restore source domain objects
        for objId in srcDom.listIds():
            if self._keyboardInterrupt:
                return False
            srcObj = srcDom.getObject(objId)
            if srcObj != None:
                if (self._restrictDomainObject != None) and (self._restrictDomainObject != srcObj.getName()):
                    continue
                self._beginDomainObjectBackup(srcObj, objectIdx, objectCount)
                objectSuccess = False
                try:
                    if self._restoreMode:                        
                        if self._restoreNewDomain:
                            self._restoreNewDomainObject = True
                            # search domain object by name and type, except for public folder                            
                            if objId == bkparser.publicFolderId:
                                destObj = destDom.getObject(objId)
                            else:
                                destObj = None
                                for destObjId in destDom.listIds():
                                    destObj = destDom.getObject(destObjId)
                                    if destObj.getName() == srcObj.getName():
                                        break
                                    destObj = None
                        else:
                            # search domain object by id
                            self._restoreNewDomainObject = False
                            destObj = destDom.getObject(objId)                            
                        if destObj == None or srcObj.getType() != destObj.getType():
                            # try to setup a different domain object in destination with an unique id
                            self._restoreNewDomainObject = True
                            destObjId = self._nextVirtualId
                            self._nextVirtualId = self._nextVirtualId + 1
                            if destDom.setObject(destObjId, srcObj.getType(), srcObj.getName()):
                                destObj = destDom.getObject(destObjId)                                
                        if destObj != None:
                            objectSuccess = self._doDomainObjectBackup(srcObj, destObj)
                        else:
                            self._reportError("Error creating domain object '{0}@{1}' in destination directory".format(srcObj.getName(), srcDom.getName()))
                    else:
                        if destDom.setObject(objId, srcObj.getType(), srcObj.getName()):
                            self._restoreNewDomainObject = False
                            destObj = destDom.getObject(objId)
                            if destObj != None:
                                objectSuccess = self._doDomainObjectBackup(srcObj, destObj)
                            else:
                                self._reportError("Destination domain object '{0}@{1}' not found".format(srcObj.getName(), srcDom.getName()))
                        else:
                            self._reportError("Error creating domain object '{0}@{1}' in destination directory".format(srcObj.getName(), srcDom.getName()))
                except KeyboardInterrupt:
                    self._keyboardInterrupt = True
                except:
                    if self._restoreMode:
                        self._reportError("Unexpected error while restoring domain object '{0}@{1}': {2} | {3}".format(srcObj.getName(), srcDom.getName(), sys.exc_info()[0], sys.exc_info()[1]))
                    else:
                        self._reportError("Unexpected error while backup domain object '{0}@{1}': {2} | {3}".format(srcObj.getName(), srcDom.getName(), sys.exc_info()[0], sys.exc_info()[1]))
                    if os.getenv('USE_DEBUG', False):
                        traceback.print_tb(sys.exc_info()[2], limit=10, file=sys.stdout)    
                self._restoreNewDomainObject = None
                objectsOk = objectsOk and objectSuccess
                self._endDomainObjectBackup(srcObj, objectSuccess, objectIdx, objectCount)
            self._reportGlobalProgress()
            objectIdx = objectIdx + 1

        if (self._restrictDomainObject == None) and objectsOk and not self._restoreMode:
            # copy object list file
            if not self._copyFile(os.path.join(srcDom.getPath(), bkparser.doListFileName), os.path.join(destDom.getPath(), bkparser.doListFileName)):
                self._reportError("Unable to copy domain objects list file for domain '{0}'".format(srcObj.getName()))
                return False
        return objectsOk

    def _doDomainObjectBackup(self, srcObj, destObj):
        if self._restrictFolder == None:
            # copy domain object config files
            for fname in listDir(srcObj.getPath()):
                if fname[-4:] == ".abf":
                    if not self._copyFile(os.path.join(srcObj.getPath(), fname), os.path.join(destObj.getPath(), fname)):
                        self._reportError("Unable to copy domain object '{0}@{1}' backup file '{2}'".format(srcObj.getName(), srcObj.backupDomain.getName(), fname))
                        return False

        if not srcObj.scan():
            if destObj.scan():
                self._reportError("Incompatible domain object types for '{0}@{1}'".format(srcObj.getName(), srcObj.backupDomain.getName()))
                return False            
            return True
        if not destObj.scan():
            self._reportError("Incompatible domain object types for '{0}@{1}'".format(srcObj.getName(), srcObj.backupDomain.getName()))
            return False            

        # start backup folders
        folderIdx = 1
        folderCount = len(srcObj.listFolderIds())
        foldersOk = True
        if self.deleteRemovedFolders:
            removedList = set(destObj.listFolderIds()) - set(srcObj.listFolderIds())
            folderCount = folderCount + len(removedList)
            for fid in removedList:
                if self._keyboardInterrupt:
                    return False
                destFolder = destObj.getFolder(fid)
                if destFolder != None:
                    if (self._restrictFolder != None) and (self._restrictFolder != destObj.getFolderPath(fid)):
                        continue
                    self._beginFolderBackup(destFolder, folderIdx, folderCount)
                    # AXI-2535 : fix for removed folders
                    if destFolder.getPath() is None:
                        destFolder.setFolderPath()
                    if os.getenv('USE_DEBUG', False):
                        print("Deleting " + str(destFolder.getPath()))
                    folderSuccess = self._deleteObject(destFolder.getPath())
                    self._endFolderBackup(destFolder, folderSuccess, folderIdx, folderCount)
                self._reportGlobalProgress(False)
                folderIdx = folderIdx + 1

        # backup source folders
        srcFolderList = set(srcObj.listFolderIds())
        while len(srcFolderList)> 0:
            retryFolderList = set()
            shouldRetry = False
            for fid in srcFolderList:
                if self._keyboardInterrupt:
                    return False
                srcFolder = srcObj.getFolder(fid)
                destFolderId = None
                if srcFolder != None:
                    if (self._restrictFolder != None) and (self._restrictFolder != srcObj.getFolderPath(fid)):
                        continue                    
                    folderSuccess = False
                    try:
                        if self._restoreMode:
                            if self._restoreNewDomainObject:
                                self._restoreNewFolder = True
                                # for new domain objects folder matching must be made by name, not by id
                                # search for a parent folder in destination
                                srcFolderPath = srcObj.getFolderPath(srcFolder.getParentId())
                                if srcFolderPath == '':
                                    destParentId = -1
                                else:
                                    destParentId = None
                                    for dfid in destObj.listFolderIds():
                                        if srcFolderPath == destObj.getFolderPath(dfid):
                                            destParentId = dfid
                                            break
                                if destParentId != None:
                                    destFolderId = self._nextVirtualId
                                    self._nextVirtualId = self._nextVirtualId + 1
                                    if destObj.setFolder(destFolderId, destParentId, srcFolder.getName()):
                                        shouldRetry = True
                                    else:
                                        destFolderId = None
                                        retryFolderList.add(fid)
                                else:
                                    retryFolderList.add(fid)
                            else:
                                # first search by source folder id
                                if destObj.getFolder(fid) != None:
                                    self._restoreNewFolder = False
                                    shouldRetry = True
                                    destFolderId = fid
                                else:
                                    # we must create a new folder in destination domain
                                    if destObj.setFolder(fid, srcFolder.getParentId(), srcFolder.getName()):
                                        self._restoreNewFolder = True
                                        shouldRetry = True
                                        destFolderId = fid
                                    else:
                                        retryFolderList.add(fid)
                        else:
                            if destObj.setFolder(fid, srcFolder.getParentId(), srcFolder.getName()):
                                self._restoreNewFolder = False
                                shouldRetry = True
                                destFolderId = fid
                            else:
                                retryFolderList.add(fid)
                    except KeyboardInterrupt:
                        self._keyboardInterrupt = True
                    except:
                        self._beginFolderBackup(srcFolder, folderIdx, folderCount)
                        self._reportError("Error creating folder '{0}+{1}@{2}' in destination directory".format(srcObj.getName(), srcObj.getFolderPath(fid), srcObj.backupDomain.getName()))
                        foldersOk = False
                        self._endFolderBackup(srcFolder, False, folderIdx, folderCount)
                        self._reportGlobalProgress(False)
                        folderIdx = folderIdx + 1

                    if destFolderId != None:
                        self._beginFolderBackup(srcFolder, folderIdx, folderCount)
                        try:
                            destFolder = destObj.getFolder(destFolderId)
                            if destFolder != None:
                                folderSuccess = self._doFolderBackup(srcFolder, destFolder)
                            else:
                                self._reportError("Destination folder '{0}+{1}@{2}' not found".format(srcObj.getName(), srcObj.getFolderPath(fid), srcObj.backupDomain.getName()))
                        except:
                            if self._restoreMode:
                                self._reportError("Unexpected error while restoring folder '{0}+{1}@{2}': {3} | {4}".format(srcObj.getName(), srcObj.getFolderPath(fid), srcObj.backupDomain.getName(), sys.exc_info()[0], sys.exc_info()[1]))
                            else:
                                self._reportError("Unexpected error while backup folder '{0}+{1}@{2}': {3} | {4}".format(srcObj.getName(), srcObj.getFolderPath(fid), srcObj.backupDomain.getName(), sys.exc_info()[0], sys.exc_info()[1]))
                            if os.getenv('USE_DEBUG', False):
                                traceback.print_tb(sys.exc_info()[2], limit=5, file=sys.stdout)    
                        srcFolder.close()
                        foldersOk = foldersOk and folderSuccess
                        self._endFolderBackup(srcFolder, folderSuccess, folderIdx, folderCount)
                else:
                    destFolderId = -1
                if destFolderId != None:
                    self._reportGlobalProgress(False)
                    folderIdx = folderIdx + 1
            self._restoreNewFolder = None
            srcFolderList = retryFolderList            
            if not shouldRetry:
                break

        for fid in srcFolderList:
            srcFolder = srcObj.getFolder(fid)
            if srcFolder != None:
                if srcObj.getFolderPath(fid) != None:
                    self._beginFolderBackup(srcFolder, folderIdx, folderCount)
                    self._reportError("Error creating folder '{0}+{1}@{2}' in destination directory".format(srcObj.getName(), srcObj.getFolderPath(fid), srcObj.backupDomain.getName()))
                    foldersOk = False
                    self._endFolderBackup(srcFolder, False, folderIdx, folderCount)
            self._reportGlobalProgress(False)
            folderIdx = folderIdx + 1
            
        if (self._restrictFolder == None) and foldersOk and not self._restoreMode:
            # copy object list file
            if not self._copyFile(os.path.join(srcObj.getPath(), bkparser.fListFileName), os.path.join(destObj.getPath(), bkparser.fListFileName)):
                self._reportError("Unable to copy folders list file for domain object '{0}@{1}'".format(srcObj.getName(), srcObj.backupDomain.getName()))
                return False
        return foldersOk

    def _doFolderBackup(self, srcFolder, destFolder):
        if self._restoreMode:
            srcFolder.setFolderPath()
        elif not srcFolder.scan():
            # not selectable folder
            return True

        destFolder.setFolderPath()
        if not self._restoreNewFolder:
            # check id validity
            if srcFolder.getIdValidity() != destFolder.getIdValidity():
                self._restoreNewFolder = True
                if not self._restoreMode:
                    # we must remove all mails from destination directory
                    try:
                        for fname in listDir(destFolder.getPath()):
                            if fname[-4:] == ".abf":
                                try:
                                    os.remove(os.path.join(destFolder.getPath(), fname))
                                except KeyboardInterrupt:
                                    self._keyboardInterrupt = True
                                    return False
                                except:
                                    pass
                    except KeyboardInterrupt:
                        self._keyboardInterrupt = True
                        return False
                    except:
                        pass
            elif (srcFolder.getModificationId() == destFolder.getModificationId()) and (srcFolder.getModificationId() != -1):
                if self._restoreMode or srcFolder.backupDomainObject.getSinceBackupId() >= destFolder.backupDomainObject.getSinceBackupId():
                    # no change in this folder
                    return True

        srcFiles = listDir(srcFolder.getPath())
        mailIdx = 1
        if self._restoreMode:
            mailCount = 0
            for fname in srcFiles:
                if fname[-4:] == ".abf" and fname[:6] != "folder":
                    mailCount = mailCount + 1
        else:
            mailCount = srcFolder.getChangesCount()
        mailsOk = True
        # copy mails and config files
        for fname in srcFiles:
            if self._keyboardInterrupt:
                return False
            if fname[-4:] == ".abf":
                if fname[:6] == "folder":
                    if not self._copyFile(os.path.join(srcFolder.getPath(), fname), os.path.join(destFolder.getPath(), fname)):
                        self._reportError("Unable to copy folder '{0}+{1}@{2}' backup file '{3}'".format(srcFolder.backupDomainObject.getName(), srcFolder.getFolderPath(), srcFolder.backupDomain.getName(), fname))
                else:
                    self._beginMailBackup(srcFolder, mailIdx, mailCount)
                    needCopyFile = True
                    if self._restoreMode and not self._restoreNewFolder:
                        # check if mail should be copied or already exists in destination
                        try:
                            if stat.S_ISREG(os.stat(os.path.join(destFolder.getPath(), fname))):
                                needCopyFile = False
                        except:
                            pass
                    if needCopyFile and not self._copyFile(os.path.join(srcFolder.getPath(), fname), os.path.join(destFolder.getPath(), fname), self.deleteRemovedMails):
                        self._reportError("Unable to copy mail {0} from folder '{1}+{2}@{3}'".format(fname[:8], srcFolder.backupDomainObject.getName(), srcFolder.getFolderPath(), srcFolder.backupDomain.getName()))
                        mailsOk = False
                        self._endMailBackup(srcFolder, False, mailIdx, mailCount)
                    else:
                        self._endMailBackup(srcFolder, True, mailIdx, mailCount)
                        self._reportGlobalProgress(False)
                    mailIdx = mailIdx + 1
        return mailsOk

    def _copyFile(self, srcFileName, destFileName, removeFile=True):
        try:
            copySize = copyFile(srcFileName, destFileName, removeFile)
        except KeyboardInterrupt:
            self._keyboardInterrupt = True
            return False
        if copySize < 0:
            return False
        self._copyFileSize = self._copyFileSize + copySize
        self._copyFileCount = self._copyFileCount + 1
        return True

    def _deleteObject(self, objectPath, removeObjectDir = True):
        try:
            return deleteObject(objectPath, removeObjectDir)
        except KeyboardInterrupt:
            self._keyboardInterrupt = True
            return False            
    
    def _beginBackup(self):
        if self.debugMode:            
            print("Begin backup: SRC='{0}' DEST='{1}'".format(self.backupSourcePath, self.backupDestinationPath))
        for r in self._reporters:
            try:
                r.beginBackup(self._srcRoot, self._destRoot, self._restoreMode)
            except KeyboardInterrupt:
                self._keyboardInterrupt = True
            except:
                pass

    def _endBackup(self, successfull):
        if self.debugMode:
            print("End backup: {0} ({1} bytes in {2} files copied)".format(successfull, self._copyFileSize, self._copyFileCount))
        for r in self._reporters:
            try:
                r.endBackup(successfull, self._copyFileCount, self._copyFileSize)
            except KeyboardInterrupt:
                self._keyboardInterrupt = True
            except:
                pass

    def _reportGlobalProgress(self, incObjIdx = True):
        if self._gobjIdx > self._gobjCount:
            self._gobjCount = self._gobjIdx
        
        for r in self._reporters:
            try:
                r.reportGlobalProgress(self._gobjIdx, self._gobjCount, self._copyFileCount, self._copyFileSize)
            except KeyboardInterrupt:
                self._keyboardInterrupt = True
            except:
                pass
        if incObjIdx:
            self._gobjIdx = self._gobjIdx + 1

    def _beginDomainBackup(self, backupDomain, domainIdx, domainCount):
        if self.debugMode:
            print("Begin backup domain '{0}'".format(backupDomain.getName()))
        if domainIdx > domainCount:
            domainCount = domainIdx
        for r in self._reporters:
            try:
                r.beginDomainBackup(backupDomain, domainIdx, domainCount)
            except KeyboardInterrupt:
                self._keyboardInterrupt = True
            except:
                pass

    def _endDomainBackup(self, backupDomain, successfull, domainIdx, domainCount):
        if self.debugMode:
            print("End backup domain '{0}': {1}".format(backupDomain.getName(), successfull))
        if domainIdx > domainCount:
            domainCount = domainIdx
        for r in self._reporters:
            try:
                r.endDomainBackup(backupDomain, successfull, domainIdx, domainCount)
            except KeyboardInterrupt:
                self._keyboardInterrupt = True
            except:
                pass

    def _abortBackup(self):
        if self.debugMode:
            print("Abort backup ({0} bytes in {1} files copied)".format(self._copyFileSize, self._copyFileCount))
        for r in self._reporters:
            try:
                r.abortBackup(self._copyFileCount, self._copyFileSize)
            except KeyboardInterrupt:
                self._keyboardInterrupt = True
            except:
                pass

    def _beginDomainObjectBackup(self, backupDomainObject, domainObjectIdx, domainObjectCount):
        if self.debugMode:
            print("Begin backup domain object '{0}@{1}'".format(backupDomainObject.getName(), backupDomainObject.backupDomain.getName()))
        if domainObjectIdx > domainObjectCount:
            domainObjectCount = domainObjectIdx
        for r in self._reporters:
            try:
                r.beginDomainObjectBackup(backupDomainObject, domainObjectIdx, domainObjectCount)
            except KeyboardInterrupt:
                self._keyboardInterrupt = True
            except:
                pass
        
    def _endDomainObjectBackup(self, backupDomainObject, successfull, domainObjectIdx, domainObjectCount):
        if self.debugMode:
            print("End backup domain object '{0}@{1}': {2}".format(backupDomainObject.getName(), backupDomainObject.backupDomain.getName(), successfull))
        if domainObjectIdx > domainObjectCount:
            domainObjectCount = domainObjectIdx
        for r in self._reporters:
            try:
                r.endDomainObjectBackup(backupDomainObject, successfull, domainObjectIdx, domainObjectCount)
            except KeyboardInterrupt:
                self._keyboardInterrupt = True
            except:
                pass

    def _beginFolderBackup(self, backupFolder, folderIdx, folderCount):
        if self.debugMode:
            print("Begin backup folder {0}+{1}@{2}'".format(backupFolder.backupDomainObject.getName(), backupFolder.backupDomainObject.getFolderPath(backupFolder.getId()), backupFolder.backupDomain.getName()))
        if folderIdx > folderCount:
            folderCount = folderIdx
        for r in self._reporters:
            try:
                r.beginFolderBackup(backupFolder, folderIdx, folderCount)
            except KeyboardInterrupt:
                self._keyboardInterrupt = True
            except:
                pass
    def _endFolderBackup(self, backupFolder, successfull, folderIdx, folderCount):
        if self.debugMode:
            print("End backup folder {0}+{1}@{2}': {3}".format(backupFolder.backupDomainObject.getName(), backupFolder.backupDomainObject.getFolderPath(backupFolder.getId()), backupFolder.backupDomain.getName(), successfull))
        if folderIdx > folderCount:
            folderCount = folderIdx
        for r in self._reporters:
            try:
                r.endFolderBackup(backupFolder, successfull, folderIdx, folderCount)
            except KeyboardInterrupt:
                self._keyboardInterrupt = True
            except:
                pass
    def _beginMailBackup(self, backupFolder, mailIdx, mailCount):
        if self.debugMode:
            print("Begin backup mail {0} from folder {1}+{2}@{3}'".format(mailIdx, backupFolder.backupDomainObject.getName(), backupFolder.backupDomainObject.getFolderPath(backupFolder.getId()), backupFolder.backupDomain.getName()))
        if mailIdx > mailCount:
            mailCount = mailIdx
        for r in self._reporters:
            try:
                r.beginMailBackup(backupFolder, mailIdx, mailCount)
            except KeyboardInterrupt:
                self._keyboardInterrupt = True
            except:
                pass
    def _endMailBackup(self, backupFolder, successfull, mailIdx, mailCount):
        if self.debugMode:
            print("End backup mail {0} from folder {1}+{2}@{3}': {4}".format(mailIdx, backupFolder.backupDomainObject.getName(), backupFolder.backupDomainObject.getFolderPath(backupFolder.getId()), backupFolder.backupDomain.getName(), successfull))
        if mailIdx > mailCount:
            mailCount = mailIdx
        for r in self._reporters:
            try:
                r.endMailBackup(backupFolder, successfull, mailIdx, mailCount)
            except KeyboardInterrupt:
                self._keyboardInterrupt = True
            except:
                pass

    def _reportError(self, errorDetails):
        if self._keyboardInterrupt:
            return
        if self.debugMode:
            print(errorDetails)
        for r in self._reporters:
            try:
                r.reportError(errorDetails)
            except KeyboardInterrupt:
                self._keyboardInterrupt = True
            except:
                pass
