
from __future__ import print_function

"""
Axigen backup parser module.
"""

import re
import os
import stat
import sys

hasTTY = False
try:
    if sys.stdout.isatty():
        hasTTY = True
except:
    pass

doListFileName = "domainObjectList.txt"
fListFileName = "folderList.txt"
fInfoFileName = "folderInfo.txt"
publicFolderId = 0xFFFFFFFD;
publicFolderType = "publicFolder"
publicFolderName = "publicFolder"
virtualObjectId = 0x100000000;

class InvalidVersionException(Exception):
    """
    Exception raised when an invalid version set for a backup file.
    """
    def __init__(self, fileVersion):
        self.fileVersion = fileVersion

class InvalidFileTypeException(Exception):
    """
    Exception raised when a backup file field list is not recognized.
    """
    def __init__(self):
        pass

class InvalidFileContentException(Exception):
    """
    Exception raised when an invalid line is read from the backup file.
    """
    def __init__(self, offendingLine):
        self.offendingLine = offendingLine

class GenericBackupException(Exception):
    """
    Exception raised when some backup error occured.
    """
    def __init__(self, description):
        self.description = description    

def renameBackupObject(parentPath, objectName):
    # find an unique name as rename destination
    bkidx = -1
    try:
        if objectName[-4] == '.':
            bkidx = int(objectName[-3:])
            objectRootName=objectName[:-3]
    except Exception:
        pass
    if bkidx < 1:
        bkidx = 0
        objectRootName = objectName + '.'
        
    oldPath = os.path.join(parentPath, objectName)
    try:
        if not stat.S_ISDIR(os.stat(oldPath).st_mode):
            return None
    except Exception:
        return None
    
    for i in range(1, 1000):
        if i == bkidx:
            continue
        newObjectName = objectRootName + "{0:03d}".format(i)
        newPath = os.path.join(parentPath, newObjectName)
        try:
            os.stat(newPath)
            continue
        except Exception:
            try:
                os.rename(oldPath, newPath)
                return newObjectName
            except Exception:
                pass
    return None
    
class BackupInfoFileParser:
    """
    Parse header part of axigen backup information files
    Recognizes lines of the form <FiledName>: <FieldValue>
    """
    def __init__(self):
        self._fields={}

    def getField(self, fieldName):
        try:
            return self._fields[fieldName]
        except Exception:
            return None

    def hasField(self, fieldName):
        return self._fields.has_key(fieldName)

    def listFields(self):
        return self._fields.keys()

    def parseFileHeader(self, backupFile):
        self._fields={}
        for line in backupFile:
            line=line.rstrip("\r\n")
            if not line:
                return
            valuePos=line.find(": ")
            if valuePos < 1:
                continue
            self._fields[line[:valuePos]] = line[valuePos+2:]

    def setField(self, fieldName, fieldValue):
        self._fields[fieldName] = fieldValue

class DomainBackupFileParser(BackupInfoFileParser):
    """
    Parse header part of axigen backup information files
    Provides methods for accessing common fileds like version, domain name, etc
    """
    def __init__(self):
        BackupInfoFileParser.__init__(self)
        
    def getVersion(self):
        try:
            return int(self.getField("Version"))
        except Exception:
            return 0

    def getDomainName(self):
        return self.getField("OwningDomain")

    def getDomainId(self):
        domainId = self.getField("DomainStorageId")
        if domainId == "invalid":
            return None
        return domainId

    def getSinceBackupId(self):
        try:
            return int(self.getField("SinceBackupId"), 16)
        except Exception:
            return 0

    def getUpdateBackupId(self):
        try:
            return int(self.getField("UpdateBackupId"), 16)
        except Exception:
            return 0
        
class DomainObjectFileParser(DomainBackupFileParser):
    """
    Parse backup domain object list files.
    """
    def __init__(self):
        DomainBackupFileParser.__init__(self)
        self._domainObjects={}
        self._parser = re.compile("(?P<id>[a-fA-F0-9]+) (?P<type>[a-zA-Z]+)/(?P<name>.*)")

    def getDomainObject(self, doId):
        try:
            return self._domainObjects[doId]
        except Exception:
            return None    

    def getDomainObjectName(self, doId):
        try:
            return self._domainObjects[doId][1]
        except Exception:
            return None

    def getDomainObjectType(self, doId):
        try:
            return self._domainObjects[doId][0]
        except Exception:
            return None

    def getDomainObjectId(self, doName):
        for doId in self._domainObjects:
            if self._domainObjects[doId][1] == doName:
                return doId
        return None

    def listIds(self):
        return self._domainObjects.keys()

    def listNames(self):
        return [self._domainObjects[objId][1] for objId in self._domainObjects]

    def parseFile(self, backupFile):
        self._domainObjects={}
        self.parseFileHeader(backupFile)
        if self.getVersion() != 1:
            raise InvalidVersionException(self.getField("Version"))
        if self.getField("FieldList") != "DomainObjectId DomainObjectFolder/DomainObjectName":
            raise InvalidFileTypeException()
        for line in backupFile:
            line=line.rstrip("\r\n")
            if not line:
                return
            m = re.match(self._parser, line)
            if not m:
                raise InvalidFileContentException(line)
            objId = int(m.group('id'), 16)
            self._domainObjects[objId] = [ m.group('type'), m.group('name') ]

    def setDomainObject(self, doId, doType, doName):
        self._domainObjects[doId] = [doType, doName]
            
class FolderContainerFileParser(DomainBackupFileParser):
    """
    Parse backup folder container list files.
    """
    def __init__(self):
        DomainBackupFileParser.__init__(self)
        self._folders={}
        self._parser = re.compile("(?P<id>[a-fA-F0-9]+) (?P<pid>[a-fA-F0-9-]+) (?P<idv>[a-fA-F0-9]+) (?P<modid>[a-fA-F0-9]+) (?P<name>.*)")

    def getFolder(self, folderId):
        try:
            return self._folders[folderId]
        except Exception:
            return None

    def listIds(self, parentId = None):
        if parentId == None:
            return self._folders.keys()
        return [fid for fid in self._folders if self._folders[fid][0] == parentId]

    def getFolderPath(self, folderId):
        if folderId == -1:
            return ''
        try:
            return self._getFolderPathRecursive(set(), folderId)[1:]
        except Exception:
            return None
    
    def getParentId(self, folderId):
        try:
            return self._folders[folderId][0]
        except Exception:
            return None

    def getIdValidity(self, folderId):
        try:
            return self._folders[folderId][1]
        except Exception:
            return None

    def getModificationId(self, folderId):
        try:
            return self._folders[folderId][2]
        except Exception:
            return None
        
    def getFolderName(self, folderId):
        try:
            return self._folders[folderId][3]
        except Exception:
            return None

    def getContainerId(self):
        return self.getField("ContainerId")

    def getOwnerId(self):
        try:
            return int(self.getField("ContainerId")[:8], 16)
        except Exception:
            return -1

    def getFolderId(self, folderPath):
        try:
            return self._getFolderIdRecursive(-1, folderPath.lower())
        except Exception:
            return None

    def setFolder(self, folderId, parentId, idValidity, modificationId, folderName):
        self._folders[folderId] = [parentId, idValidity, modificationId, folderName]

    def setParentId(self, folderId, parentId):
        self._folders[folderId][0] = parentId

    def setFolderName(self, folderId, folderName):
        self._folders[folderId][3] = folderName
    
    def parseFile(self, backupFile):
        self._folders={}
        self.parseFileHeader(backupFile)
        if self.getVersion() != 1:
            raise InvalidVersionException(self.getField("Version"))
        if self.getField("FieldList") != "MboxId ParentId IdValidity ModificationId MboxName.MboxType":
            raise InvalidFileTypeException()
        for line in backupFile:
            line=line.rstrip("\r\n")
            if not line:
                return
            m = re.match(self._parser, line)
            if not m:
                raise InvalidFileContentException(line)
            mboxId = int(m.group('id'), 16)
            if m.group('pid')[0] == '-':
                parentId = -1
            else:
                parentId = int(m.group('pid'), 16)
            idValidity = int(m.group('idv'), 16)
            modId = int(m.group('modid'), 16)
            mboxName = m.group('name')
            self._folders[mboxId] = [ parentId, idValidity, modId, mboxName ]

    def _getFolderPathRecursive(self, fList, folderId):
        if folderId == -1:
            return ""
        if fList & set([folderId]):
            return None
        fList.add(folderId)
        crFolder = self._folders[folderId]
        return self._getFolderPathRecursive(fList, crFolder[0]) + "/" + crFolder[3]

    def _getFolderIdRecursive(self, parentId, folderName):
        if not folderName:
            return parentId
        if folderName[0] == '/':
            return self._getFolderIdRecursive(parentId, folderName[1:])
        nextNamePos = folderName.find('/')
        if nextNamePos < 1:
            crName = folderName
            nextNamePos = len(folderName) - 1
        else:
            crName = folderName[:nextNamePos]
        idList = self.listIds(parentId)
        if idList == None:
            return None
        for fid in idList:
            fname = self.getFolderName(fid).lower()
            if fname == crName:
                return self._getFolderIdRecursive(fid, folderName[nextNamePos + 1:])
            dotPos = fname.rfind('.')
            if (dotPos > 0) and (fname[:dotPos] == crName):
                return self._getFolderIdRecursive(fid, folderName[nextNamePos + 1:])
            
        return None
    
class FolderInfoFileParser(BackupInfoFileParser):
    """
    Parse folder information backup files.
    """
    def __init__(self):
        BackupInfoFileParser.__init__(self)

    def getContainerId(self):
        return self.getField("ContainerId")

    def getOwnerId(self):
        try:
            return int(self.getField("ContainerId")[:8], 16)
        except Exception:
            return -1

    def getFolderId(self):
        try:
            return int(self.getField("MboxId"), 16)
        except Exception:
            return -1

    def getChangesCount(self):
        try:
            return int(self.getField("ChangesCount"))
        except Exception:
            return -1

    def getModificationId(self):
        try:
            return int(self.getField("ModificationId"), 16)
        except Exception:
            return -1

    def parseFile(self, backupFile):
        self.parseFileHeader(backupFile)

class BackupFolder:
    """
    Contains backup information about a folder
    """
    def __init__(self, backupRoot, backupDomain, backupDomainObject, ownerPath, folderId, foldersList, folderObj):
        self.backupRoot = backupRoot
        self.backupDomain = backupDomain
        self.backupDomainObject = backupDomainObject
        self._ownerPath = ownerPath
        self._folderId = folderId
        self._folders = foldersList
        self._folderObj = folderObj
        self._folderInfo = FolderInfoFileParser()
        self._infoFile = None
        self._folderPath = None

    def scan(self):
        try:
            folderPath=self._folders.getFolderPath(self._folderId)
            if folderPath == None:
                raise GenericBackupException("Folder path not found for " + self._folderObj[3])
            try:
                self._infoFile = open(os.path.join(self._ownerPath, folderPath, fInfoFileName), "r")
            except IOError:
                return False
            self._folderInfo.parseFile(self._infoFile)
            if self._folderInfo.getContainerId() != self._folders.getContainerId():
                raise InvalidFileContentException("Container id does not match for folder " + self._folderObj[3])
            if (self._folderInfo.getFolderId() != self._folderId) and (self._folderId < virtualObjectId):
                raise InvalidFileContentException("Folder id does not match for " + self._folderObj[3])
            self._folderPath = self._ownerPath + "/" + folderPath
        except:
            self.close()
            raise
        return True

    def setFolderPath(self):
        folderPath=self._folders.getFolderPath(self._folderId)
        if folderPath == None:
            raise GenericBackupException("Folder path not found for " + self._folderObj[3])
        self._folderPath = self._ownerPath + "/" + folderPath
        
    def close(self):
        if self._infoFile != None:
            try:
                self._infoFile.close()
            except Exception:
                pass
            self._infoFile = None

    def getPath(self):
        return self._folderPath
    
    def getName(self):
        return self._folderObj[3]

    def getId(self):
        return self._folderId

    def getParentId(self):
        return self._folderObj[0]

    def getIdValidity(self):
        return self._folderObj[1]
    
    def getModificationId(self):
        if self._folderPath != None:
            modId = self._folderInfo.getModificationId()
            if modId >= 0:
                return modId
        return self._folderObj[2]

    def getFolderPath(self):
        return self.backupDomainObject.getFolderPath(self._folderId)

    def listSubfolders(self):
        return self._folders.listIds(self._folderId)

    def getFolder(self, folderId):
        folderObj = self._folders.getFolder(folderId)
        if folderObj == None:
            return None
        return BackupFolder(self.backupRoot, self.backupDomain, self.backupDomainObject, self._ownerPath, folderId, self._folders, folderObj)

    def getContainerId(self):
        return self._folders.getContainerId()

    def getOwnerId(self):
        return self._folders.getOwnerId()

    def getChangesCount(self):
        return self._folderInfo.getChangesCount()
    
class BackupDomainObject:
    """
    Contains backup information about a domain object
    """
    def __init__(self, backupRoot, backupDomain, objectPath, objectId, objectName, objectType):
        self.backupRoot = backupRoot
        self.backupDomain = backupDomain
        self._objectPath = objectPath
        self._objectId = objectId
        self._objectName = objectName
        self._objectType = objectType
        self._folders = FolderContainerFileParser()

    def scan(self):
        if self._objectType != "accounts" and self._objectType != publicFolderType:
            return False
        try:
            with open(os.path.join(self._objectPath, fListFileName), "r") as doFile:
                self._folders.parseFile(doFile)
            if (self._folders.getOwnerId() != self._objectId) and (self._objectId < virtualObjectId):
                raise InvalidFileContentException("Container id does not match for object " + self._objectName)
        except IOError:
            pass
        return True

    def getPath(self):
        return self._objectPath

    def getName(self):
        return self._objectName

    def getType(self):
        return self._objectType

    def getId(self):
        return self._objectId

    def getContainerId(self):
        return self._folders.getContainerId()

    def listFolderIds(self, parentId = None):
        return self._folders.listIds(parentId)

    def getFolder(self, folderId):
        folderObj = self._folders.getFolder(folderId)
        if folderObj == None:
            return None
        return BackupFolder(self.backupRoot, self.backupDomain, self, self._objectPath, folderId, self._folders, folderObj)    

    def getFolderName(self, folderId):
        return self._folders.getFolderName(folderId)

    def getFolderPath(self, folderId):
        return self._folders.getFolderPath(folderId)

    def getFolderParentId(self, folderId):
        return self._folders.getParentId(folderId)

    def getFolderIdValidity(self, folderId):
        return self._folders.getIdValidity(folderId)

    def getFolderModificationId(self, folderId):
        return self._folders.getModificationId(folderId)

    def getFolderId(self, folderPath):
        return self._folders.getFolderId(folderPath)

    def getSinceBackupId(self):
        return self._folders.getSinceBackupId()

    def getUpdateBackupId(self):
        return self._folders.getUpdateBackupId()    

    def setFolder(self, folderId, parentId, folderName):
        parentFolderPath = self.getFolderPath(parentId)
        if parentFolderPath == None:
            return False
        if (self.getFolderParentId(folderId) == parentId) and (self.getFolderName(folderId) == folderName):
            return True
        # need to create or rename the folder
        # make sure destination directory is available
        parentPath = os.path.join(self._objectPath, parentFolderPath)
        if folderId < virtualObjectId:
            newName = renameBackupObject(parentPath, folderName)
            if newName:
                oldId = self.getFolderId(parentFolderPath + "/" + folderName)
                if oldId != None:
                    self._folders.setFolderName(oldId, newName)
                    if not hasTTY:
                        print('Rename-F1 from %s to %s' % (folderName,newName))
        newPath = os.path.join(parentPath, folderName)
        try:
            srcFolderPath = self.getFolderPath(folderId)
            if srcFolderPath == None:
                # need to create a new folder
                if (folderId < virtualObjectId) or not os.path.isdir(newPath):
                    os.makedirs(newPath)
                self._folders.setFolder(folderId, parentId, 0, 0, folderName)
            else:
                # need to rename the folder
                if not hasTTY:
                    print('Rename-F2 from %s to %s' % (srcFolderPath,folderName))
                os.rename(os.path.join(self._objectPath, srcFolderPath), newPath)
                self._folders.setParentId(folderId, parentId)
                self._folders.setFolderName(folderId, folderName)
        except Exception as e:
            if not hasTTY:
                print('Exception: ' + str(e))
            return False
        return True
            
class BackupDomain:
    """
    Contains backup information about a domain
    """
    def __init__(self, backupRoot, domainPath, domainObjects):
        self.backupRoot = backupRoot
        self._domainPath = domainPath
        self._domainObjects = domainObjects

    def scan(self):
        self._domainObjects.setDomainObject(publicFolderId, publicFolderType, publicFolderName)
        try:
            publicFolderPath = os.path.join(self._domainPath, publicFolderType)
            if not os.path.isdir(publicFolderPath):
              os.mkdir(publicFolderPath)
        except Exception as e:
            if not hasTTY:
                print('Exception: ' + str(e))
            pass
        return True
    
    def getPath(self):
        return self._domainPath

    def getName(self):
        return self._domainObjects.getDomainName()

    def getId(self):
        return self._domainObjects.getDomainId()
    
    def getSinceBackupId(self):
        return self._domainObjects.getSinceBackupId()

    def getUpdateBackupId(self):
        return self._domainObjects.getUpdateBackupId()

    def listNames(self):
        return self._domainObjects.listNames()

    def listIds(self):
        return self._domainObjects.listIds()

    def getObjectName(self, objId):
        return self._domainObjects.getDomainObjectName(objId)

    def getObjectId(self, objName):
        return self._domainObjects.getDomainObjectId(objName)

    def getObject(self, objId):
        obj = self._domainObjects.getDomainObject(objId)
        if obj == None:
            return None
        if objId == publicFolderId:
            return BackupDomainObject(self.backupRoot, self, os.path.join(self._domainPath, obj[0]), objId, obj[1], obj[0])
        return BackupDomainObject(self.backupRoot, self, os.path.join(self._domainPath, obj[0], obj[1]), objId, obj[1], obj[0])

    def getObjectType(self, objId):
        return self._domainObjects.getDomainObjectType(objId)

    def setObject(self, objId, objType, objName):
        newPath = os.path.join(self._domainPath, objType, objName)
        if self.getObjectName(objId) == objName:
            if self.getObjectType(objId) == objType:
                return True
        elif self.getObjectType(objId) == objType:
            # handle object rename
            if not hasTTY:
                print('Rename-O1 from %s to %s' % (self.getObjectName(objId), newPath))
            try:
                os.rename(os.path.join(self._domainPath, objType, self.getObjectName(objId)), newPath)
                self._domainObjects.setDomainObject(objId, objType, objName)
                return True
            except Exception as e:
                if not hasTTY:
                    print('Exception: ' + str(e))
                return False
        # we must create a new object
        newName = renameBackupObject(os.path.join(self._domainPath, objType), objName)
        if newName:
            oldObjId = self.getObjectId(objName)
            if oldObjId != None:
                self._domainObjects.setDomainObject(oldObjId, objType, newName)
        try:
            os.makedirs(newPath)
            self._domainObjects.setDomainObject(objId, objType, objName)
        except Exception as e:
            if not hasTTY:
                print('Exception: ' + str(e))
            return False
        return True

class BackupRoot:
    """
    Scan a root backup directory for backup domains.
    """
    def __init__(self, rootPath):
        self._rootPath = rootPath
        self._domains = {}

    def scan(self):
        subdirs = os.listdir(self._rootPath)
        for dname in subdirs:
            try:
                with open(os.path.join(self._rootPath, dname, doListFileName), 'r') as doFile:
                    dofp = DomainObjectFileParser()
                    dofp.parseFile(doFile)
                    self._domains[dname] = dofp
            except Exception:
                pass
        return True

    def listNames(self):
        return self._domains.keys()

    def listIds(self):
        return [self._domains[crDomain].getDomainId() for crDomain in self._domains]

    def listDomainObjectsCount(self):
        return [len(self._domains[crDomain].listIds()) for crDomain in self._domains]

    def getDomainName(self, domainId):
        for dname in self._domains.keys():
            if self._domains[dname].getDomainId() == domainId:
                return dname
        return None

    def getDomainId(self, domainName):
        try:
            return self._domains[domainName].getDomainId()
        except Exception:
            return None
            
    def getDomainByName(self, domainName):
        try:
            return BackupDomain(self, os.path.join(self._rootPath, domainName), self._domains[domainName])
        except Exception:
            return None

    def getDomainById(self, domainId):
        for dname in self._domains.keys():
            if self._domains[dname].getDomainId() == domainId:
                return BackupDomain(self, os.path.join(self._rootPath, dname), self._domains[dname])
        return None
        
    def getPath(self):
        return self._rootPath
    
    def setDomain(self, domainId, domainName):
        if domainName in self._domains.keys():
            # the normal situation
            if self._domains[domainName].getDomainId() == domainId:
                return True
            # another domain with a different id found - rename the current domain
            newName = renameBackupObject(self._rootPath, domainName)
            if not newName:
                # unable to rename old domain
                if not hasTTY:
                    print('Rename-D1 failed for %s' % (domainName))
                return False
            if not hasTTY:
                print('Rename-D1 from %s to %s' % (domainName,newName))
            oldDomainObj = self._domains[domainName]
            del self._domains[domainName]
            oldDomainObj._domainPath = os.path.join(self._rootPath, newName)
            self._domains[newName] = oldDomainObj
        else:
            newName = renameBackupObject(self._rootPath, domainName)
            if newName:
                if not hasTTY:
                    print('Rename-D2 from %s to %s' % (domainName,newName))
        newPath = os.path.join(self._rootPath, domainName)
        # check if domainId already used by some domain
        for dname in self._domains.keys():
            if self._domains[dname].getDomainId() == domainId:
                # rename the domain
                if not hasTTY:
                    print('Rename-D3 from %s to %s' % (dname,domainName))
                try:
                    os.rename(os.path.join(self._rootPath, dname), newPath)
                    oldDomainObj = self._domains[dname]
                    del self._domains[dname]
                    oldDomainObj._domainPath = newPath
                    self._domains[domainName] = oldDomainObj
                    return True
                except Exception as e:
                    if not hasTTY:
                        print('Exception: ' + str(e))
                    return False
        # need to create a new domain folder and object
        try:
            os.mkdir(newPath)
            dofp = DomainObjectFileParser()
            dofp.setField("OwningDomain", domainName)
            dofp.setField("DomainStorageId", domainId)
            self._domains[domainName] = dofp
        except Exception as e:
            if not hasTTY:
                print('Exception: ' + str(e))
            return False
        return True
